#include <iostream>
#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <CGAL/Delaunay_triangulation_2.h>
#include <CGAL/Triangulation_hierarchy_2.h>
#include <CGAL/point_generators_2.h>
#include <CGAL/iterator.h>
#include <CGAL/Timer.h>
//typedef CGAL::Simple_cartesian<double> K;
typedef CGAL::Exact_predicates_inexact_constructions_kernel K;
typedef CGAL::Delaunay_triangulation_2<K>                   DT;
typedef CGAL::Triangulation_vertex_base_2<K>                Vbb;
typedef CGAL::Triangulation_hierarchy_vertex_base_2<Vbb>    Vb;
typedef CGAL::Triangulation_face_base_2<K>                  Fb;
typedef CGAL::Triangulation_data_structure_2<Vb,Fb>         Tds;
typedef CGAL::Delaunay_triangulation_2<K,Tds>               Dt;
typedef CGAL::Triangulation_hierarchy_2<Dt>                 HDT;
typedef CGAL::Random_points_in_square_2<DT::Point>          Rnd_points_it;

void compute_delaunay_no_hint(std::vector<DT::Point>::iterator it, 
		      std::vector<DT::Point>::iterator e){
    HDT dt;
    for( ;it!=e; ++it)  dt.insert(*it);
    return;
}
void compute_delaunay(std::vector<DT::Point>::iterator it, 
		      std::vector<DT::Point>::iterator e){
    DT dt;
    DT::Face_handle hint;
    for( ;it!=e; ++it)  hint = dt.insert(*it, hint)->face();
    return;
}

int main (int argc, char **argv)
{   int N = 1000000; if( argc > 1 )N = atoi(argv[1]); // number of points
    Rnd_points_it   rand_it( 1.0);
    std::vector<DT::Point> points;
    for (int i=0; i< N; ++i) {
      points.push_back( *rand_it++ );
    }

    DT dt;
    CGAL::Timer cost;

    std::cout << " Delaunay triangulation of "<<N<<" random points in dim 2"<< std::endl;

    cost.reset();cost.start();
    std::cout << "  Delaunay without  sort...                "<< std::flush;
    compute_delaunay(points.begin(), points.end());cost.stop();
    std::cout << "done in "<<cost.time()<<" seconds." << std::endl;

    cost.reset();cost.start();
    std::cout << "  Delaunay with hierarchy...                "<< std::flush;
    compute_delaunay_no_hint(points.begin(), points.end());cost.stop();
    std::cout << "done in "<<cost.time()<<" seconds." << std::endl;

    cost.reset();cost.start();
    std::cout << "  Delaunay with x sort...                "<< std::flush;
    std::sort(points.begin(), points.end() );
    std::cout << "sort in "<<cost.time()<<" seconds." << std::flush;
    cost.reset();
    compute_delaunay(points.begin(), points.end());cost.stop();
    std::cout << "done in "<<cost.time()<<" seconds." << std::endl;

    cost.reset();cost.start();
    std::cout << "  Delaunay with Hilbert sort (median policy)...   " << std::flush;
    CGAL::hilbert_sort(points.begin(), points.end(), CGAL::Hilbert_sort_median_policy());
    std::cout << "sort in "<<cost.time()<<" seconds." << std::flush;
    cost.reset();
    compute_delaunay(points.begin(), points.end());cost.stop();
    std::cout << "done in "<<cost.time()<<" seconds." << std::endl;

    cost.reset();cost.start();
    std::cout << "  Delaunay with Hilbert sort (middle policy)...   " << std::flush;
    CGAL::hilbert_sort(points.begin(), points.end(), CGAL::Hilbert_sort_middle_policy() );
    std::cout << "sort in "<<cost.time()<<" seconds." << std::flush;
    cost.reset();
    compute_delaunay(points.begin(), points.end());cost.stop();
    std::cout << "done in "<<cost.time()<<" seconds." << std::endl;

    cost.reset();cost.start();
    std::cout << "  Delaunay with spatial sort (median policy)...   " << std::flush;
    CGAL::spatial_sort(points.begin(), points.end(), CGAL::Hilbert_sort_median_policy());
    std::cout << "sort in "<<cost.time()<<" seconds." << std::flush;
    cost.reset();
    compute_delaunay(points.begin(), points.end());cost.stop();
    std::cout << "done in "<<cost.time()<<" seconds." << std::endl;

    cost.reset();cost.start();
    std::cout << "  Delaunay with spatial sort (middle policy)...   " << std::flush;
    CGAL::spatial_sort(points.begin(), points.end(), CGAL::Hilbert_sort_middle_policy() );
    std::cout << "sort in "<<cost.time()<<" seconds." << std::flush;
    cost.reset();
    compute_delaunay(points.begin(), points.end());cost.stop();
    std::cout << "done in "<<cost.time()<<" seconds." << std::endl;


    std::vector<DT::Point> parab;
    N/=10;
    std::cout << " Delaunay triangulation of "<<N<<" parabola points in dim 2"<< std::endl;
    for (int i=0; i< N; ++i) {
      double x= -N +i;
      parab.push_back( DT::Point( x, x*x ));
    }

    cost.reset();cost.start();
    std::cout << "  Delaunay in x order...                "<< std::flush;
    compute_delaunay(parab.begin(), parab.end());cost.stop();
    std::cout << "done in "<<cost.time()<<" seconds." << std::endl;

    cost.reset();cost.start();
    std::cout << "  Delaunay with random order...                "<< std::flush;
    std::random_shuffle(parab.begin(), parab.end() );
    std::cout << "shuffle in "<<cost.time()<<" seconds." << std::flush;
    cost.reset();
    compute_delaunay(parab.begin(), parab.end());cost.stop();
    std::cout << "done in "<<cost.time()<<" seconds." << std::endl;

    cost.reset();cost.start();
    std::cout << "  Delaunay with hierarchy...                "<< std::flush;
    compute_delaunay_no_hint(parab.begin(), parab.end());cost.stop();
    std::cout << "done in "<<cost.time()<<" seconds." << std::endl;

    cost.reset();cost.start();
    std::cout << "  Delaunay with Hilbert sort (median policy)...   " << std::flush;
    CGAL::hilbert_sort(parab.begin(), parab.end(), CGAL::Hilbert_sort_median_policy());
    std::cout << "sort in "<<cost.time()<<" seconds." << std::flush;
    cost.reset();
    compute_delaunay(parab.begin(), parab.end());cost.stop();
    std::cout << "done in "<<cost.time()<<" seconds." << std::endl;

    cost.reset();cost.start();
    std::cout << "  Delaunay with Hilbert sort (middle policy)...   " << std::flush;
    CGAL::hilbert_sort(parab.begin(), parab.end(), CGAL::Hilbert_sort_middle_policy() );
    std::cout << "sort in "<<cost.time()<<" seconds." << std::flush;
    cost.reset();
    compute_delaunay(parab.begin(), parab.end());cost.stop();
    std::cout << "done in "<<cost.time()<<" seconds." << std::endl;

    cost.reset();cost.start();
    std::cout << "  Delaunay with spatial sort (median policy)...   " << std::flush;
    CGAL::spatial_sort(parab.begin(), parab.end(), CGAL::Hilbert_sort_median_policy());
    std::cout << "sort in "<<cost.time()<<" seconds." << std::flush;
    cost.reset();
    compute_delaunay(parab.begin(), parab.end());cost.stop();
    std::cout << "done in "<<cost.time()<<" seconds." << std::endl;

    cost.reset();cost.start();
    std::cout << "  Delaunay with spatial sort (middle policy)...   " << std::flush;
    CGAL::spatial_sort(parab.begin(), parab.end(), CGAL::Hilbert_sort_middle_policy() );
    std::cout << "sort in "<<cost.time()<<" seconds." << std::flush;
    cost.reset();
    compute_delaunay(parab.begin(), parab.end());cost.stop();
    std::cout << "done in "<<cost.time()<<" seconds." << std::endl;





    return 0;
}
